# `@trusted-pdf/pdfa`

## Пример использования
### Browser
```ts
import { PDFA } from "@trusted-pdf/pdfa";
import { DOMCanvasFactory } from "pdfjs-dist/types/src/display/display_utils";

const pdfa = new PDFA({
  canvasFactory: new DOMCanvasFactory(),
});

const data = await pdfa.convert(file);
```

### NodeJs
```ts
import * as fs from "node:fs";
import { PDFA } from "@trusted-pdf/pdfa";
import { NodeCanvasFactory } from "./node_canvas_factory";

const file = fs.readFileSync(`path/to/input.pdf`);

const pdfa = new PDFA({
  canvasFactory: new NodeCanvasFactory(),
  url: {
    cMapUrl: `${__dirname}/node_modules/pdfjs-dist/cmaps/`,
    standardFontDataUrl: `${__dirname}/node_modules/pdfjs-dist/standard_fonts/`,
  }
});

const data = await pdfa.convert(file);

fs.writeFileSync(`path/to/output.pdf`, data);
```
## Примеры работы со шрифтами.

```ts
const file = fs.readFileSync(`path/to/input.pdf`);

const pdfa = new PDFA({
  canvasFactory: new DOMCanvasFactory(),
});

// Получение списка незарегистрированных шрифтов.
const unregisteredFonts = pdfa.fonts.findUnregisteredFonts(file);

// Регистрация шрифта Tahoma.
pdfa.fonts.add(FontExtNames.Tahoma, new Uint8Array(fs.readFileSync(`${__dirname}/../testdata/Tahoma.ttf`)));

// Получение списка шрифтов.
const fontList = pdfa.fonts.list();

// Удаление шрифта Arial.
pdfa.fonts.delete(FontExtNames.Arial);

const data = await pdfa.convert(file);
```
### Список дополнительных стандартных шрифтов не описанных в spec.
  - Tahoma,
  - Tahoma-Bold,
  - ArialNarrow,
  - ArialNarrow-Bold,
  - ArialNarrow-BoldItalic,
  - ArialNarrow-Italic,
  - ArialBlack,
  - ArialBlack-Bold,
  - ArialBlack-BoldItalic,
  - ArialBlack-Italic,
  - Arial-Black,
  - Arial-Black-Bold,
  - Arial-Black-BoldItalic,
  - Arial-Black-Italic,
  - Arial,
  - Arial-Bold,
  - Arial-BoldItalic,
  - Arial-Italic,
  - ArialMT,
  - Arial-BoldItalicMT,
  - Arial-BoldMT,
  - Arial-ItalicMT,
  - Arial-BoldItalicMT-BoldItalic,
  - Arial-BoldMT-Bold,
  - Arial-ItalicMT-Italic,
  - ArialUnicodeMS,
  - ArialUnicodeMS-Bold,
  - ArialUnicodeMS-BoldItalic,
  - ArialUnicodeMS-Italic,
  - Courier-BoldItalic,
  - Courier-Italic,
  - CourierNew,
  - CourierNew-Bold,
  - CourierNew-BoldItalic,
  - CourierNew-Italic,
  - CourierNewPS-BoldItalicMT,
  - CourierNewPS-BoldMT,
  - CourierNewPS-ItalicMT,
  - CourierNewPSMT,
  - Helvetica-BoldItalic,
  - Helvetica-Italic,
  - Symbol-Bold,
  - Symbol-BoldItalic,
  - Symbol-Italic,
  - TimesNewRoman,
  - TimesNewRoman-Bold,
  - TimesNewRoman-BoldItalic,
  - TimesNewRoman-Italic,
  - TimesNewRomanPS,
  - TimesNewRomanPS-Bold,
  - TimesNewRomanPS-BoldItalic,
  - TimesNewRomanPS-BoldItalicMT,
  - TimesNewRomanPS-BoldMT,
  - TimesNewRomanPS-Italic,
  - TimesNewRomanPS-ItalicMT,
  - TimesNewRomanPSMT,
  - TimesNewRomanPSMT-Bold,
  - TimesNewRomanPSMT-BoldItalic,
  - TimesNewRomanPSMT-Italic,
